# Copyright (c) Meta Platforms, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

if __name__ == "__main__":
    unittest.main("fixit.tests", verbosity=2)
