# Copyright (c) Meta Platforms, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Linting framework built on LibCST, with automatic fixes
"""

__version__ = "0.2.0"

from .api import fixit_bytes, fixit_file, fixit_paths
from .types import Config, FileContent, Result

__all__ = [
    "__version__",
]
