# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from fixit.common.base import CstLintRule


# Dummy rule for integration testing purposes.
class DummyRule1(CstLintRule):
    pass
