# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
For backwards compatibility.
"""

from fixit.cli.formatter import LintRuleReportFormatter


__all__ = ["LintRuleReportFormatter"]
